@@ -74,6 +74,7 @@ int main(int argc, char *argv[])
 	int brief = 0;
 	int force = 0;
 	int test = 0;
+	int export = 0;
 	int assume_clean = 0;
 	char *symlinks = NULL;
 	/* autof indicates whether and how to create device node.
@@ -106,6 +107,7 @@ int main(int argc, char *argv[])
 
 	int copies;
 	int print_help = 0;
+	FILE *outf;
 
 	int mdfd = -1;
 
@@ -158,6 +160,9 @@ int main(int argc, char *argv[])
 			}
 			continue;
 
+		case 'Y': export++;
+			continue;
+
 		case HomeHost:
 			homehost = optarg;
 			continue;
@@ -193,6 +198,8 @@ int main(int argc, char *argv[])
 		case 'F': newmode = MONITOR;break;
 		case 'G': newmode = GROW; shortopt = short_bitmap_auto_options; break;
 		case 'I': newmode = INCREMENTAL; break;
+		case AutoDetect:
+			newmode = AUTODETECT; break;
 
 		case '#':
 		case 'D':
@@ -226,7 +233,7 @@ int main(int argc, char *argv[])
 			if (opt == 'c' && 
 			    ( strncmp(optarg, "--h", 3)==0 ||
 			      strncmp(optarg, "-h", 2)==0)) {
-				fputs(Help_config, stderr);
+				fputs(Help_config, stdout);
 				exit(0);
 			}
 
@@ -272,6 +279,7 @@ int main(int argc, char *argv[])
 		case 'F':
 		case 'G':
 		case 'I':
+		case AutoDetect:
 			continue;
 		}
 		if (opt == 1) {
@@ -604,17 +612,24 @@ int main(int argc, char *argv[])
 
 				continue;
 			}
-			if (strcmp(update,"?") == 0 || strcmp(update, "help") == 0)
-				fprintf(stderr, Name ": ");
-			else
-				fprintf(stderr, Name ": '--update=%s' is invalid.  ", update);
-			fprintf(stderr, "Valid --update options are:\n"
+			if (strcmp(update,"?") == 0 ||
+			    strcmp(update, "help") == 0) {
+				outf = stdout;
+				fprintf(outf, Name ": ");
+			} else {
+				outf = stderr;
+				fprintf(outf,
+					Name ": '--update=%s' is invalid.  ",
+					update);
+			}
+			fprintf(outf, "Valid --update options are:\n"
 		"     'sparc2.2', 'super-minor', 'uuid', 'name', 'resync',\n"
 		"     'summaries', 'homehost', 'byteorder', 'devicesize'.\n");
-			exit(2);
+			exit(outf == stdout ? 0 : 2);
 
 		case O(ASSEMBLE,NoDegraded): /* --no-degraded */
-			runstop = -1; /* --stop isn't allowed for --assemble, so we overload slightly */
+			runstop = -1; /* --stop isn't allowed for --assemble,
+				       * so we overload slightly */
 			continue;
 
 		case O(ASSEMBLE,'c'): /* config file */
@@ -873,7 +888,7 @@ int main(int argc, char *argv[])
 			case GROW     : help_text = Help_grow; break;
 			case INCREMENTAL:help_text= Help_incr; break;
 			}
-		fputs(help_text,stderr);
+		fputs(help_text,stdout);
 		exit(0);
 	}
 
@@ -1126,6 +1141,11 @@ int main(int argc, char *argv[])
 			rv = 1;
 			break;
 		}
+		if (raiddisks == 0) {
+			fprintf(stderr, Name ": no raid-disks specified.\n");
+			rv = 1;
+			break;
+		}
 
 		if (bitmap_file) {
 			if (strcmp(bitmap_file, "internal")==0) {
@@ -1145,6 +1165,11 @@ int main(int argc, char *argv[])
 			rv = 1;
 			break;
 		}
+		if (raiddisks == 0) {
+			fprintf(stderr, Name ": no raid-disks specified.\n");
+			rv = 1;
+			break;
+		}
 
 		rv = Create(ss, devlist->devname, mdfd, chunk, level, layout, size<0 ? 0 : size,
 			    raiddisks, sparedisks, ident.name, homehost,
@@ -1181,7 +1206,8 @@ int main(int argc, char *argv[])
 								e->dev);
 							continue;
 						}
-						rv |= Detail(name, verbose>1?0:verbose+1, test, homehost);
+						rv |= Detail(name, verbose>1?0:verbose+1,
+							     export, test, homehost);
 						put_md_name(name);
 					}
 				} else	if (devmode == 'S' && scan) {
@@ -1226,7 +1252,10 @@ int main(int argc, char *argv[])
 			for (dv=devlist ; dv; dv=dv->next) {
 				switch(dv->disposition) {
 				case 'D':
-					rv |= Detail(dv->devname, brief?1+verbose:0, test, homehost); continue;
+					rv |= Detail(dv->devname,
+						     brief?1+verbose:0,
+						     export, test, homehost);
+					continue;
 				case 'K': /* Zero superblock */
 					rv |= Kill(dv->devname, force, quiet); continue;
 				case 'Q':
@@ -1328,6 +1357,10 @@ int main(int argc, char *argv[])
 		}
 		rv = Incremental(devlist->devname, verbose-quiet, runstop,
 				 ss, homehost, autof);
+		break;
+	case AUTODETECT:
+		autodetect();
+		break;
 	}
 	exit(rv);
 }
